import time
import uuid
import os
import datetime
from threading import Lock

from taos.cinterface import IS_V3
from taos.constants import FieldType
from taos.precision import PrecisionEnum, PrecisionError
from taos import log
from taos.error import *


tUUIDHashId = 0
tUUIDSerialNo = 0


def gen_req_id():
    global tUUIDHashId
    global tUUIDSerialNo
    if tUUIDHashId == 0:
        uid = str(uuid.uuid4()).encode("utf-8")
        tUUIDHashId = murmurhash3_32(uid, len(uid))

    while True:
        ts = int(time.time()) >> 8
        pid = get_pid()
        with Lock():
            tUUIDSerialNo += 1
            val = tUUIDSerialNo

        req_id = ((tUUIDHashId & 0x07FF) << 52) | ((pid & 0x0F) << 48) | ((ts & 0x3FFFFFF) << 20) | (val & 0xFFFFF)
        if req_id:
            break

    return req_id


def murmurhash3_32(key, length):
    data = key
    nblocks = length >> 2
    h1 = 0x12345678
    c1 = 0xCC9E2D51
    c2 = 0x1B873593
    blocks = data
    for i in range(-nblocks, 0):
        k1 = blocks[i]
        k1 *= c1
        k1 = rotl32(k1, 15)
        k1 *= c2
        h1 ^= k1
        h1 = rotl32(h1, 13)
        h1 = h1 * 5 + 0xE6546B64
    tail = data[nblocks * 4 :]
    k1 = 0
    # print('length', length, length & 3)
    if length & 3 == 3:
        k1 ^= tail[2] << 16
    if length & 3 == 2:
        k1 ^= tail[1] << 8
    if length & 3 == 1:
        k1 ^= tail[0]
        k1 *= c1
        k1 = rotl32(k1, 15)
        k1 *= c2
        h1 ^= k1
    h1 ^= length
    fmix32(h1)
    return h1


def rotl32(x, r):
    return (x << r) | (x >> (32 - r))


def fmix32(h):
    h ^= h >> 16
    h *= 0x85EBCA6B
    h ^= h >> 13
    h *= 0xC2B2AE35
    h ^= h >> 16
    return h


def get_pid():
    return os.getpid()


#
# detect list contian none count
#
ALL_NONE = 0  # all list item is none
HAVE_NONE = 1  # list item some is none, some is not none
NOFOUND_NONE = 2  # all list item no found none


def detectListNone(items):
    if items is None:
        return ALL_NONE
    # loop
    n0 = 0  # none count
    n1 = 0  # not noen count
    for item in items:
        if item is None:
            n0 += 1
        else:
            n1 += 1

    # return
    if n0 == 0:
        return NOFOUND_NONE
    if n1 == 0:
        return ALL_NONE
    return HAVE_NONE


def checkString(label, values, types):
    for value in values:
        if value is None:
            continue
        # type
        if type(value) not in types:
            err = f"{label} type bind not support type = {type(value)}"
            raise DataTypeAndRangeError(err)
        # check length should do for engine
        # if length is not None and len(value) + 2 > length:
        #    err = f"{label} type value:{value} length exceeds the max length {length}"
        #    raise DataTypeAndRangeError(err)


def checkNumber(label, values, types, min, max):
    for value in values:
        if value is None:
            continue
        # type
        if type(value) not in types:
            err = f"{label} type bind not support type: {type(value)}"
            raise DataTypeAndRangeError(err)
        # range
        if value < min or value > max:
            err = f"{label} type value:{value} exceeds the indicated range [{min}, {max}]"
            raise DataTypeAndRangeError(err)


def checkTypeValid(buffer_type, values):
    if buffer_type == FieldType.C_TIMESTAMP:
        for value in values:
            if value is None:
                continue
            # check type
            if type(value) not in (int, str, datetime.datetime):
                err = f"timestamp type bind not support type = {type(value)}"
                raise DataTypeAndRangeError(err)
            elif type(value) is int:
                # check range (same bigint)
                min = -(2**63)
                max = 2**63 - 1
                if value < min or value > max:
                    err = f"timestamp type value:{value} exceeds the indicated range [{min}, {max}]"
                    raise DataTypeAndRangeError(err)
    elif buffer_type == FieldType.C_BOOL:
        for value in values:
            if value is None:
                continue
            # check type
            if type(value) not in (bool, int, float):
                err = f"bool type bind not support type = {type(value)}"
                raise DataTypeAndRangeError(err)
    elif buffer_type == FieldType.C_TINYINT:
        checkNumber("tinyint", values, [int, float], -128, 127)
    elif buffer_type == FieldType.C_SMALLINT:
        checkNumber("smallint", values, [int, float], -32768, 32767)
    elif buffer_type == FieldType.C_INT:
        checkNumber("int", values, [int, float], -(2**31), 2**31 - 1)
    elif buffer_type == FieldType.C_BIGINT:
        checkNumber("bigint", values, [int, float], -(2**63), 2**63 - 1)
    elif buffer_type == FieldType.C_FLOAT:
        checkNumber("float", values, [int, float], -3.4e38, 3.4e38)
    elif buffer_type == FieldType.C_DOUBLE:
        checkNumber("double", values, [int, float], -1.7e308, 1.7e308)
    # unsigned
    elif buffer_type == FieldType.C_TINYINT_UNSIGNED:
        checkNumber("unsigned tinyint", values, [int, float], 0, 255)
    elif buffer_type == FieldType.C_SMALLINT_UNSIGNED:
        checkNumber("unsigned smallint", values, [int, float], 0, 65535)
    elif buffer_type == FieldType.C_INT_UNSIGNED:
        checkNumber("unsigned int", values, [int, float], 0, 2**32 - 1)
    elif buffer_type == FieldType.C_BIGINT_UNSIGNED:
        checkNumber("unsigned bigint", values, [int, float], 0, 2**64 - 1)
    # str
    elif buffer_type == FieldType.C_VARCHAR:
        checkString("varchar", values, [str])
    elif buffer_type == FieldType.C_BINARY:
        checkString("binary", values, [str])
    elif buffer_type == FieldType.C_NCHAR:
        checkString("nchar", values, [str])
    elif buffer_type == FieldType.C_JSON:
        checkString("json", values, [str])
    elif buffer_type == FieldType.C_VARBINARY:
        checkString("varbinary", values, [bytes, bytearray])
    elif buffer_type == FieldType.C_GEOMETRY:
        checkString("geometry", values, [bytes, bytearray])
    elif buffer_type == FieldType.C_BLOB:
        checkString("blob", values, [bytes, bytearray])
    else:
        err = f"invalid datatype type={buffer_type} values= {values}"
        raise DataTypeAndRangeError(err)



import base64
import datetime
import json
from urllib.request import urlopen, Request

import requests

from iso8601 import parse_date
from pytz import timezone as Timezone
from typing import Union, Optional

from .errors import ConnectError, ExecutionError, HTTPError

error_msgs = {
    400: "parameter error",
    401: "authorization error",
    404: "api not found",
    500: "internal error",
    503: "system resources is not sufficient. It is may be caused by a huge query.",
}


class RestClient:
    """
    A wrapper for TDengine REST API.
    For detailed info about TDengine REST API refer https://docs.tdengine.com/reference/rest-api/
    """

    def __init__(
        self,
        url: str,
        token: str = None,
        database: str = None,
        user: str = "root",
        password: str = "taosdata",
        timeout: int = None,
        convert_timestamp: bool = True,
        timezone: Union[str, datetime.tzinfo] = None,
    ):
        """
        Create a RestClient object.

        Parameters
        -----------
        - url : service address, required. for example: https://192.168.1.103:6041
        - token : cloud service token, optional
        - database: default database to use
        - user : username used to log in, optional
        - password : password used to log in, optional
        - timeout : the optional timeout parameter specifies a timeout in seconds for blocking operations
        - convert_timestamp: whether to convert timestamp from type str to type datetime with tzinfo.
            The default timezone is UTC. You can use method `datatime.astimezone()` to convert it to your local time.
        - timezone: str | datetime.tzinfo, optional, default None.
            When convert_timestamp is true, which timezone to used.
            When the type of timezone is str, it should be recognized by [pytz package](https://pypi.org/project/pytz/).
            When the timezone is None, system timezone will be used and the returned datetime object will be offset-naive (no tzinfo), otherwise the returned datetime will be offset-aware(with tzinfo)
        """
        # determine schema://host:post
        self._url = url.strip("/")
        if not self._url.startswith("http://") and not self._url.startswith("https://"):
            self._url = "http://" + self._url
        # timeout
        self._timeout = timeout
        # determine full URL to use and the header to user.
        if token:
            if not database:
                self._sql_url = f"{self._url}/rest/sql?token={token}"
            else:
                self._sql_url = f"{self._url}/rest/sql/{database}?token={token}"
            self._headers = {}
        else:
            if not database:
                self._sql_url = f"{self._url}/rest/sql"
            else:
                self._sql_url = f"{self._url}/rest/sql/{database}"

            original_str = f"{user}:{password}"
            original_bytes = original_str.encode("utf-8")
            encoded_bytes = base64.b64encode(original_bytes)
            auth = encoded_bytes.decode("utf-8")
            self._headers = {"Authorization": "Basic " + auth}

        self._convert_timestamp = convert_timestamp

        if timezone is None:
            self._timezone = None
        elif isinstance(timezone, str):
            self._timezone = Timezone(timezone)
        elif isinstance(timezone, datetime.tzinfo):
            self._timezone = timezone
        else:
            raise TypeError("timezone argument must be an instance of str or tzinfo")

    def sql(self, q: str, req_id: Optional[int] = None) -> dict:
        """
        Execute sql and return the JSON response.

        If http status not equals to 200 or "code" in response not equals to 0, then Error will be raised.

        Parameters
        -----------
        q : SQL statement to execute. Can't be USE statement since REST api is stateless.
        req_id : request id, optional
        """
        # print(f"execute rest sql = {q}\n")
        data = q.encode("utf8")
        if req_id:
            url = f"{self._sql_url}?req_id={req_id}"
        else:
            url = self._sql_url

        r = requests.post(url, data=data, headers=self._headers, timeout=self._timeout)

        if not r.ok:
            raise HTTPError(r.status_code, error_msgs.get(r.status_code, "unknown error"))

        resp = r.json()

        if "status" in resp:
            # v2
            if resp["status"] != "succ":
                raise ConnectError(resp["desc"], status=resp["status"])
        else:
            if resp["code"] != 0:
                raise ConnectError(resp["desc"], resp["code"])
        if self._convert_timestamp:
            self._convert_time(resp)
        return resp

    def _convert_time(self, resp: dict):
        """
        Convert timestamp in string format to python's datetime object with time zone info.
        """
        meta = resp["column_meta"]
        data = resp["data"]
        ts_cols = []
        for i in range(len(meta)):
            if meta[i][1] == "TIMESTAMP" or meta[i][1] == 9:
                ts_cols.append(i)
        if len(ts_cols) == 0:
            return
        if self._timezone:
            for row in data:
                for i in ts_cols:
                    if row[i]:
                        dt = parse_date(row[i])  # UTC
                        row[i] = dt.astimezone(self._timezone)
        else:
            for row in data:
                for i in ts_cols:
                    if row[i]:
                        dt = parse_date(row[i])  # UTC
                        dt = dt.astimezone()  # local
                        row[i] = dt.replace(tzinfo=None)  # naive datetime

    def _check_status(self, response):
        status = response.status
        if status != 200:
            msg = error_msgs.get(status)
            raise HTTPError(status, msg)



# info
show_info = True
show_warn = True
show_debug = False
show_debug1 = False
show_debug2 = False
show_debug3 = False


# set show
def setting(info, warn, debug, debug1, debug2, debug3):
    global show_info, show_warn, show_debug, show_debug1, show_debug2, show_debug3
    show_info = info
    show_warn = warn
    show_debug = debug
    show_debug1 = debug1
    show_debug2 = debug2
    show_debug3 = debug3


# info
def info(msg):
    if show_info:
        print(msg)


# warn
def warn(msg):
    if show_warn:
        print(msg)


# debug
def debug(msg):
    if show_debug:
        print(msg)


# debug1
def debug1(msg):
    if show_debug1:
        print(msg)


# debug2
def debug2(msg):
    if show_debug2:
        print(msg)


# debug3
def debug3(msg):
    if show_debug3:
        print(msg)



from typing import List, Dict
from .errors import NotSupportedError, ConnectError
from .cursor import TaosRestCursor
from .restclient import RestClient
from typing import Optional


class Result:
    def __init__(self, resp: dict):
        if "status" in resp:
            self.status: str = resp["status"]
        else:
            self.code: str = resp["code"]
        self.column_meta: List[list] = resp["column_meta"]
        self.data: List[list] = resp["data"]
        self.rows: int = resp["rows"]

    @property
    def field_count(self):
        return len(self.column_meta)

    @property
    def fields(self) -> List[Dict]:
        """
        return a list of column meta dict which contains three keys:
            - name: column name
            - type: column type code
            - bytes: data length in bytes
        for more information about column meta,
         refer https://docs.tdengine.com/2.4/reference/rest-api/#http-return-format
        """
        return list(map(lambda meta: {"name": meta[0], "type": meta[1], "bytes": meta[2]}, self.column_meta))

    def __iter__(self):
        return self.data.__iter__()


class TaosRestConnection:
    """
    Implement [PEP 249 Connection API](https://peps.python.org/pep-0249/#connection-objects)
    """

    default_configs = {"url", "token", "user", "password", "database", "timeout", "convert_timestamp", "timezone"}

    def __init__(self, **kwargs):
        """
        Keyword Arguments
        ----------------------------
        - url: str, optional, default "http://localhost:6041"
             url to connect
        - token: str, optional, default None
             TDengine cloud Token, which is required only by TDengine cloud service
        - user : str, optional, default root
            username used to log in
        - password : str, optional, default taosdata
            password used to log in
        - database : str, optional, default None
             default database to use.
        - timeout : int, optional.
            the optional timeout parameter specifies a timeout in seconds for blocking operations
        - convert_timestamp: bool, optional, default true
             whether to convert timestamp in RFC3339 format to python datatime.
         - timezone: str | datetime.tzinfo, optional, default None.
             When convert_timestamp is true, which timezone to used.
             When the type of timezone is str, it should be recognized by
             [pytz package](https://pypi.org/project/pytz/).
             When the timezone is None, system timezone will be used and the returned datetime object will be
             offset-naive (no tzinfo), otherwise the returned datetime will be offset-aware(with tzinfo)
        """
        for key in kwargs:
            if key not in self.default_configs:
                raise ConnectError("Unrecognized configs: %s" % key)

        self._url = kwargs.get("url", "http://localhost:6041")
        self._token = kwargs.get("token")
        self._user = kwargs.get("user", "root")
        self._password = kwargs.get("password", "taosdata")
        self._database = kwargs.get("database")
        self._timeout = kwargs.get("timeout")
        self._convert_timestamp = kwargs.get("convert_timestamp", True)
        self._timezone = kwargs.get("timezone", None)
        self._client = RestClient(
            self._url,
            token=self._token,
            database=self._database,
            user=self._user,
            password=self._password,
            timeout=self._timeout,
            convert_timestamp=self._convert_timestamp,
            timezone=self._timezone,
        )

    def close(self):
        pass

    def commit(self):
        pass

    def rollback(self):
        raise NotSupportedError()

    def cursor(self):
        return TaosRestCursor(self._client)

    ############################################################################
    # Methods bellow are not PEP249 specified.
    # Add them for giving a similar programming experience as taos.Connection.
    ############################################################################

    @property
    def server_info(self):
        resp = self._client.sql("select server_version()")
        if len(resp["data"]) > 0:
            return resp["data"][0][0]
        return ""

    def execute(self, sql: str, req_id: Optional[int] = None) -> Optional[int]:
        """
        execute none query statement and return affected row count.
        If there is not a column named "affected_rows" in response, then None is returned.
        """
        resp = self._client.sql(sql, req_id=req_id)
        if resp["column_meta"][0][0] == "affected_rows":
            return resp["data"][0][0]
        return None

    def query(self, sql: str, req_id: Optional[int] = None) -> Result:
        """
        execute sql and wrap the http response as Result object.
        """
        resp = self._client.sql(sql, req_id=req_id)
        return Result(resp)



from sqlalchemy import sql
from sqlalchemy import text
from sqlalchemy import types as sqltypes
from sqlalchemy.engine import default, reflection
from sqlalchemy.sql import compiler

TYPES_MAP = {
    "BOOL": sqltypes.Boolean,
    "TIMESTAMP": sqltypes.DATETIME,
    "INT": sqltypes.Integer,
    "INT UNSIGNED": sqltypes.Integer,
    "BIGINT": sqltypes.BigInteger,
    "BIGINT UNSIGNED": sqltypes.BigInteger,
    "FLOAT": sqltypes.FLOAT,
    "DOUBLE": sqltypes.FLOAT,
    "TINYINT": sqltypes.SmallInteger,
    "TINYINT UNSIGNED": sqltypes.SmallInteger,
    "SMALLINT": sqltypes.SmallInteger,
    "SMALLINT UNSIGNED": sqltypes.SmallInteger,
    "BINARY": sqltypes.String,
    "VARCHAR": sqltypes.String,
    "VARBINARY": sqltypes.BINARY,
    "NCHAR": sqltypes.Unicode,
    "JSON": sqltypes.JSON,
    "BLOB": sqltypes.BLOB,
    "GEOMETRY": sqltypes.BINARY,
}

# TDengine reserved words
RESERVED_WORDS_TDENGINE = {
    "account",
    "accounts",
    "add",
    "aggregate",
    "all",
    "alter",
    "analyze",
    "and",
    "anti",
    "anode",
    "anodes",
    "anomaly_window",
    "apps",
    "as",
    "asc",
    "asof",
    "at_once",
    "balance",
    "batch_scan",
    "between",
    "bigint",
    "binary",
    "bnode",
    "bnodes",
    "bool",
    "both",
    "buffer",
    "bufsize",
    "by",
    "cache",
    "cachemodel",
    "cachesize",
    "case",
    "cast",
    "child",
    "client_version",
    "cluster",
    "column",
    "comment",
    "comp",
    "compact",
    "compacts",
    "connection",
    "connections",
    "conns",
    "consumer",
    "consumers",
    "contains",
    "count",
    "count_window",
    "create",
    "createdb",
    "current_user",
    "database",
    "databases",
    "dbs",
    "decimal",
    "delete",
    "delete_mark",
    "desc",
    "describe",
    "distinct",
    "distributed",
    "dnode",
    "dnodes",
    "double",
    "drop",
    "duration",
    "else",
    "enable",
    "encryptions",
    "encrypt_algorithm",
    "encrypt_key",
    "end",
    "exists",
    "expired",
    "explain",
    "event_window",
    "every",
    "file",
    "fill",
    "fill_history",
    "first",
    "float",
    "flush",
    "from",
    "for",
    "force",
    "full",
    "function",
    "functions",
    "geometry",
    "grant",
    "grants",
    "full",
    "logs",
    "machines",
    "group",
    "hash_join",
    "having",
    "host",
    "if",
    "ignore",
    "import",
    "in",
    "index",
    "indexes",
    "inner",
    "insert",
    "int",
    "integer",
    "interval",
    "into",
    "is",
    "jlimit",
    "join",
    "json",
    "keep",
    "key",
    "kill",
    "language",
    "last",
    "last_row",
    "leader",
    "leading",
    "left",
    "licences",
    "like",
    "limit",
    "linear",
    "local",
    "match",
    "maxrows",
    "max_delay",
    "bwlimit",
    "merge",
    "meta",
    "only",
    "minrows",
    "minus",
    "mnode",
    "mnodes",
    "modify",
    "modules",
    "normal",
    "nchar",
    "next",
    "near",
    "nmatch",
    "none",
    "not",
    "now",
    "no_batch_scan",
    "null",
    "null_f",
    "nulls",
    "offset",
    "on",
    "or",
    "order",
    "outer",
    "outputtype",
    "pages",
    "pagesize",
    "para_tables_sort",
    "partition",
    "partition_first",
    "pass",
    "port",
    "position",
    "pps",
    "primary",
    "precision",
    "prev",
    "privileges",
    "qnode",
    "qnodes",
    "qtime",
    "queries",
    "query",
    "pi",
    "rand",
    "range",
    "ratio",
    "pause",
    "read",
    "recursive",
    "redistribute",
    "rename",
    "replace",
    "replica",
    "reset",
    "resume",
    "restore",
    "retentions",
    "revoke",
    "right",
    "rollup",
    "schemaless",
    "scores",
    "select",
    "semi",
    "server_status",
    "server_version",
    "session",
    "set",
    "show",
    "single_stable",
    "skip_tsma",
    "sliding",
    "slimit",
    "sma",
    "smalldata_ts_sort",
    "smallint",
    "snode",
    "snodes",
    "sort_for_group",
    "soffset",
    "split",
    "stable",
    "stables",
    "start",
    "state",
    "state_window",
    "storage",
    "stream",
    "streams",
    "strict",
    "stt_trigger",
    "subscribe",
    "subscriptions",
    "substr",
    "substring",
    "subtable",
    "sysinfo",
    "system",
    "table",
    "tables",
    "table_prefix",
    "table_suffix",
    "tag",
    "tags",
    "tbname",
    "then",
    "timestamp",
    "timezone",
    "tinyint",
    "to",
    "today",
    "topic",
    "topics",
    "trailing",
    "transaction",
    "transactions",
    "trigger",
    "trim",
    "tsdb_pagesize",
    "tseries",
    "tsma",
    "tsmas",
    "ttl",
    "union",
    "unsafe",
    "unsigned",
    "untreated",
    "update",
    "use",
    "user",
    "users",
    "using",
    "value",
    "value_f",
    "values",
    "varchar",
    "variables",
    "verbose",
    "vgroup",
    "vgroups",
    "view",
    "views",
    "vnode",
    "vnodes",
    "wal_fsync_period",
    "wal_level",
    "wal_retention_period",
    "wal_retention_size",
    "wal_roll_period",
    "wal_segment_size",
    "watermark",
    "when",
    "where",
    "window",
    "window_close",
    "window_offset",
    "with",
    "write",
    "_c0",
    "_irowts",
    "_irowts_origin",
    "_isfilled",
    "_qduration",
    "_qend",
    "_qstart",
    "_rowts",
    "_tags",
    "_wduration",
    "_wend",
    "_wstart",
    "_flow",
    "_fhigh",
    "_frowts",
    "alive",
    "varbinary",
    "s3_chunkpages",
    "s3_keeplocal",
    "s3_compact",
    "s3migrate",
    "keep_time_offset",
    "arbgroups",
    "is_import",
    "force_window_close",
}

# backup generator function
"""
generator from TDengine/source/libs/parse/src/parTokenizer.c -> keywordTable

import sys
def readKeyWord(filename):
    keys = ""
    print(f"read file {filename}\n")
    with open(filename) as file:
        for line in file.readlines():
            pos1 = line.find('"')
            if pos1 == -1 :
                print(f"NO FOUND FIRST QUOTA: {line}\n")
                continue
            pos2 = line.find('"', pos1 + 1)
            if pos2 == -1 :
                print(f"NO FOUND SECOND QUOTA: {line}\n")
                continue
            word = line[pos1:pos2+1]
            if keys == "":
                keys = "RESERVED_WORDS_TDENGINE = {\n    " + word.lower()
            else:
                keys += ",\n    " + word.lower()

    # end
    keys += "\n}"
    print(f"\n\n{keys}\n")


if __name__ == "__main__":
    readKeyWord("./keyword.txt")

"""


class TDengineCompiler(compiler.SQLCompiler):
    """TDengine SQL compiler with simplified output format"""

    def visit_create_table(self, create, **kw):
        """Custom CREATE TABLE statement format"""
        table = create.element
        preparer = self.preparer

        # Build basic CREATE TABLE statement
        text = "CREATE TABLE " + preparer.format_table(table)

        # Build column definitions
        create_column_spec = self.get_column_specification
        columns = []
        for column in table.columns:
            column_spec = self.create_column_specification(column)
            columns.append(column_spec)

        # Use compact format
        text += " (" + ", ".join(columns) + ")"

        return text

    def create_column_specification(self, column, **kwargs):
        """Create column definition with simplified format"""
        spec = self.preparer.format_column(column)
        spec += " " + self.dialect.type_compiler.process(column.type)
        return spec


class TDengineDDLCompiler(compiler.DDLCompiler):
    def visit_create_table(self, create, **kw):
        """Simplified CREATE TABLE format"""
        return super().visit_create_table(create, **kw).replace("\n\t", " ").replace("\n", " ")


#
# identifier for TDengine
#
class TDengineIdentifierPreparer(sql.compiler.IdentifierPreparer):
    reserved_words = RESERVED_WORDS_TDENGINE

    def __init__(self, dialect, server_ansiquotes=False, **kw):
        if not server_ansiquotes:
            quote = "`"
        else:
            quote = '"'

        super(TDengineIdentifierPreparer, self).__init__(dialect, initial_quote=quote, escape_quote=quote)

    def _quote_free_identifiers(self, *ids):
        """Unilaterally identifier-quote any number of strings."""
        return tuple([self.quote_identifier(i) for i in ids if i is not None])


#
# base class for dialect
#
class BaseDialect(default.DefaultDialect):
    supports_native_boolean = True
    implicit_returning = True
    # supports_statement_cache = True

    # Set back-quote identifier preparer for TDengine keywords
    preparer = TDengineIdentifierPreparer

    def is_sys_db(self, dbname):
        return dbname.lower() in ["information_schema", "performance_schema"]

    def do_rollback(self, connection):
        pass

    def _get_server_version_info(self, connection):
        cursor = connection.execute(text("select server_version()"))
        return cursor.fetchone()

    def do_execute(self, cursor, statement, parameters, context=None):
        if parameters is None or len(parameters) == 0:
            cursor.execute(statement, parameters)
        else:
            cursor.execute(statement, [parameters])

        return cursor

    def do_executemany(self, cursor, statement, parameters, context=None):
        cursor.executemany(statement, parameters)
        return cursor

    @reflection.cache
    def has_schema(self, connection, schema, **kw):
        return schema in self.get_schema_names(connection)

    # Check if table exists
    @reflection.cache
    def has_table(self, connection, table_name, schema=None, **kw):
        return table_name in self.get_table_names(connection, schema)

    # Get column information
    @reflection.cache
    def get_columns(self, connection, table_name, schema=None, **kw):
        sysdb = False
        if schema is None:
            sql = f"describe {table_name}"
        else:
            sql = f"describe {schema}.{table_name}"
            # sysdb = self.is_sys_db(schema)
        try:
            cursor = connection.execute(text(sql))
            columns = []
            for row in cursor.fetchall():
                # print(row)
                column = dict()
                column["name"] = row[0]
                column["type"] = self._resolve_type(row[1])
                columns.append(column)
            return columns
        except:
            return []

    @reflection.cache
    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        columns = self.get_columns(connection, table_name, schema)
        return {"constrained_columns": [columns[0]["name"]], "name": None}

    @reflection.cache
    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        # No foreign key is supported by TDengine
        return []

    # Get indexes information
    @reflection.cache
    def get_indexes(self, connection, table_name, schema=None, **kw):
        sql = (
            "SELECT * FROM information_schema.INS_INDEXES "
            f"WHERE db_name = '{schema}'"
            f"AND table_name = '{table_name}'"
        )
        try:
            cursor = connection.execute(text(sql))
            rows = cursor.fetchall()
            indexes = []
            for row in rows:
                index = {"name": row[0], "column_names": [row[5]], "type": "index", "unique": False}
                indexes.append(index)
            return indexes
        except:
            return []

    # Get database names
    @reflection.cache
    def get_schema_names(self, connection, **kw):
        sql = text("SHOW DATABASES")
        try:
            cursor = connection.execute(sql)
            names = []
            for row in cursor.fetchall():
                if self.is_sys_db(row[0]) is False:
                    names.append(row[0])
            return names
        except:
            return []

    # Get table names
    @reflection.cache
    def get_table_names(self, connection, schema=None, **kw):
        if schema is None:
            sqls = [f"show stables", f"show normal tables"]
        else:
            sqls = [f"show `{schema}`.stables", f"show normal `{schema}`.tables"]
        # Execute queries
        try:
            names = []
            for sql in sqls:
                cursor = connection.execute(text(sql))
                for row in cursor.fetchall():
                    names.append(row[0])
            return names
        except:
            return []

    @reflection.cache
    def get_view_names(self, connection, schema=None, **kw):
        if schema is None:
            return []
        # SQL query for views
        sql = f"show `{schema}`.views"
        # Execute query
        try:

            cursor = connection.execute(text(sql))
            return [row[0] for row in cursor.fetchall()]
        except:
            return []

    def _resolve_type(self, type_):
        # print(f"call function {sys._getframe().f_code.co_name} type: {type_} ...\n")
        return TYPES_MAP.get(type_, sqltypes.UserDefinedType)


#
# ---------------- TDengine native connector implementation -------------
#
import taos


# TDengine native dialect
class TaosDialect(BaseDialect):
    name = "taos"
    driver = "taos"
    supports_statement_cache = True
    statement_compiler = TDengineCompiler
    ddl_compiler = TDengineDDLCompiler

    @classmethod
    def dbapi(cls):
        return taos

    @classmethod
    def import_dbapi(cls):
        return taos


#
# ---------------- TDengine WebSocket connector implementation -------------
#


# WebSocket dialect
class TaosWsDialect(BaseDialect):
    name = "taosws"
    driver = "taosws"

    @classmethod
    def dbapi(cls):
        import taosws

        return taosws

    @classmethod
    def import_dbapi(cls):
        import taosws

        return taosws

    @classmethod
    def create_connect_args(cls, url):
        if url.username and url.password:
            userpass = f"{url.username}:{url.password}"
        elif url.username:
            userpass = f"{url.username}"
        elif url.password:
            userpass = f":{url.password}"
        else:
            userpass = ""

        at = "@" if userpass else ""

        hosts = url.query.get("hosts")
        if hosts:
            addr = hosts
        else:
            if url.host and url.port:
                addr = f"{url.host}:{url.port}"
            elif url.host:
                addr = f"{url.host}"
            elif url.port:
                addr = f":{url.port}"
            else:
                addr = ""

        params = ""
        for i, (key, value) in enumerate(url.query.items()):
            if key == "hosts":
                continue
            params += f"{key}={value}"
            if i != len(url.query.items()) - 1:
                params += "&"

        dsn = f"{url.drivername}://{userpass}{at}{addr}"
        if url.database:
            dsn += f"/{url.database}"
        if params:
            dsn += f"?{params}"

        return ([dsn], {})



# encoding:UTF-8
from typing import Optional

from taos import log
from taos.cinterface import *
from taos.constants import FieldType
from taos.error import *


class TaosCursor(object):
    """Database cursor which is used to manage the context of a fetch operation.

    Attributes:
        .description: Read-only attribute consists of 7-item sequences:

            > name (mandatory)
            > type_code (mandatory)
            > display_size
            > internal_size
            > precision
            > scale
            > null_ok

            This attribute will be None for operations that do not return rows or
            if the cursor has not had an operation invoked via the .execute*() method yet.

        .rowcount:This read-only attribute specifies the number of rows that the last
            .execute*() produced (for DQL statements like SELECT) or affected
    """

    def __init__(self, connection=None, decode_binary=True):
        self._description = []
        self._rowcount = -1
        self._connection = None
        self._result = None
        self._stmt_result = None
        self._fields = None
        self._block = None
        self._stmt = None
        self._bind_sql = ""
        self._block_rows = -1
        self._block_iter = 0
        self._affected_rows = 0
        self._logfile = ""
        self.decode_binary = decode_binary

        if connection is not None:
            self._connection = connection

    def __iter__(self):
        return self

    def __next__(self):
        return self._taos_next()

    def next(self):
        return self._taos_next()

    def _taos_next(self):
        if self._result is None or self._fields is None:
            raise OperationalError("Invalid use of fetch iterator")

        if self._block_rows <= self._block_iter:
            block, self._block_rows = taos_fetch_row(self._result, self._fields, decode_binary=self.decode_binary)
            if self._block_rows == 0:
                raise StopIteration
            self._block = list(map(tuple, zip(*block)))
            self._block_iter = 0

        data = self._block[self._block_iter]
        self._block_iter += 1

        return data

    @property
    def description(self):
        """Return the description of the object."""
        return self._description

    @property
    def rowcount(self):
        """
        For INSERT statement, rowcount is assigned immediately after execute the statement.
        For SELECT statement, rowcount will not get correct value until fetched all data.
        """
        return self._rowcount

    @property
    def affected_rows(self):
        """Return the rowcount of insertion"""
        return self._affected_rows

    def callproc(self, procname, *args):
        """Call a stored database procedure with the given name.

        Void functionality since no stored procedures.
        """
        pass

    def log(self, logfile):
        self._logfile = logfile

    def close(self):
        """Close the cursor."""
        if self._connection is None:
            return False
        if self._stmt is not None:
            self._stmt.close()
            self._stmt = None
        self._reset_result()
        self._connection = None
        self._bind_sql = ""

        return True

    def keys(self):
        """Return the list of column names"""
        if self._fields:
            return self._fields
        elif self._stmt_result and self._stmt_result.fields:
            return self._stmt_result.fields
        return []

    def execute(self, operation, params=None, req_id: Optional[int] = None):
        if not operation:
            return None
        log.debug(f"execute: {operation} with params: {params}")
        self._reset_result()
        if params is not None and isinstance(params, (dict, list, tuple)) and len(params) > 0:
            return self._execute_stmt(operation, params)
        else:
            return self._execute_sql(operation, req_id=req_id)

    def _execute_stmt(self, operation, params, is_convert=False):
        """Prepare and execute a database operation (query or command)."""

        if not self._connection:
            # TODO : change the exception raised here
            raise ProgrammingError("Cursor is not connected")

        if self._stmt is None or self._bind_sql != operation:
            if self._stmt is not None:
                self._stmt.close()
                self._stmt = None

            self._stmt = self._connection.statement2(operation)
            self._bind_sql = operation
            log.debug(f"bind sql: {operation}, params: {params}")
            if self._stmt is None:
                raise OperationalError("Failed to initialize statement")
            # Here we handle the data format of unbound super tables called by SQLAlchemy
            if is_convert and not self._stmt.is_tbname:
                if isinstance(params[0], (list, tuple, set)):
                    params = [[list(col) for col in zip(*params)]]

        self._stmt.bind_param(None, None, params)
        self._stmt.execute()
        if self._stmt.is_insert():
            self._affected_rows = self._stmt.affected_rows
            self._rowcount = self._affected_rows
            return self._affected_rows
        else:
            self._stmt_result = self._stmt.result()
            self._fields = self._stmt_result.fields
            self._handle_result()
            return self._stmt_result

    def _execute_sql(self, operation, req_id: Optional[int] = None):
        """Prepare and execute a database operation (query or command)."""

        if not self._connection:
            raise ProgrammingError("Cursor is not connected")

        sql = operation

        if req_id is None:
            self._result = taos_query(self._connection._conn, sql)
        else:
            self._result = taos_query_with_reqid(self._connection._conn, sql, req_id)
        # print("   << Query ({}) Exec Done".format(localSeqNum))
        if self._logfile:
            with open(self._logfile, "a", encoding="utf-8") as logfile:
                logfile.write("%s;\n" % operation)

        if taos_field_count(self._result) == 0:
            affected_rows = taos_affected_rows(self._result)
            self._affected_rows = affected_rows
            self._rowcount = affected_rows
            return affected_rows
        else:
            self._fields = taos_fetch_fields(self._result)
            return self._handle_result()

    def executemany(self, operation, data_list):
        """
        Prepare a database operation (query or command) and then execute it against all parameter sequences or mappings
        found in the sequence seq_of_parameters.
        """
        if not operation or not data_list or len(data_list) == 0:
            return None

        self._reset_result()

        return self._execute_stmt(operation, data_list, True)

    def execute_many(self, operation, data_list, req_id: Optional[int] = None):
        """
        Prepare a database operation (query or command) and then execute it against all parameter sequences or mappings
        found in the sequence seq_of_parameters.
        """
        sql = operation
        flag = True
        affected_rows = 0
        for line in data_list:
            if isinstance(line, dict):
                flag = False
                affected_rows += self.execute(sql.format(**line), req_id=req_id)
            elif isinstance(line, list):
                sql += f" {tuple(line)} "
            elif isinstance(line, tuple):
                sql += f" {line} "
        if flag:
            affected_rows += self.execute(sql, req_id=req_id)
        return affected_rows

    def fetchone(self):
        """Fetch the next row of a query result set, returning a single sequence, or None when no more data is available."""
        try:
            if self._stmt_result is not None:
                return self._stmt_result.next()
            else:
                return self._taos_next()
        except StopIteration:
            return None

    def fetchmany(self, size: int = None):
        """Fetch the next set of rows of a query result, returning a sequence of sequences."""
        if self._result is None or self._fields is None:
            raise OperationalError("Invalid use of fetch iterator")
        if size is None:
            return self.fetchall()

        if size <= 0:
            raise ValueError("size must be greater than 0")

        data = []
        count = 0

        try:
            while count < size:
                # If current block is exhausted, fetch new data block
                if self._block_iter >= self._block_rows:
                    block, self._block_rows = taos_fetch_block(
                        self._result, self._fields, decode_binary=self.decode_binary
                    )

                    # If no more data available, break the loop
                    if self._block_rows == 0:
                        break

                    self._block = list(map(tuple, zip(*block)))
                    self._block_iter = 0

                # Calculate how many rows can be taken from current block
                remaining_in_block = self._block_rows - self._block_iter
                needed = size - count
                rows_to_take = min(remaining_in_block, needed)

                # Get data from current block
                end_index = self._block_iter + rows_to_take
                data.extend(self._block[self._block_iter : end_index])

                # Update counters
                self._block_iter += rows_to_take
                count += rows_to_take

        except Exception as e:
            errno = taos_errno(self._result)
            if errno != 0:
                raise ProgrammingError(taos_errstr(self._result), errno)
            raise e

        return data

    def istype(self, col, dataType):
        if dataType.upper() == "BOOL":
            if self._description[col][1] == FieldType.C_BOOL:
                return True
        if dataType.upper() == "TINYINT":
            if self._description[col][1] == FieldType.C_TINYINT:
                return True
        if dataType.upper() == "TINYINT UNSIGNED":
            if self._description[col][1] == FieldType.C_TINYINT_UNSIGNED:
                return True
        if dataType.upper() == "SMALLINT":
            if self._description[col][1] == FieldType.C_SMALLINT:
                return True
        if dataType.upper() == "SMALLINT UNSIGNED":
            if self._description[col][1] == FieldType.C_SMALLINT_UNSIGNED:
                return True
        if dataType.upper() == "INT":
            if self._description[col][1] == FieldType.C_INT:
                return True
        if dataType.upper() == "INT UNSIGNED":
            if self._description[col][1] == FieldType.C_INT_UNSIGNED:
                return True
        if dataType.upper() == "BIGINT":
            if self._description[col][1] == FieldType.C_BIGINT:
                return True
        if dataType.upper() == "BIGINT UNSIGNED":
            if self._description[col][1] == FieldType.C_BIGINT_UNSIGNED:
                return True
        if dataType.upper() == "FLOAT":
            if self._description[col][1] == FieldType.C_FLOAT:
                return True
        if dataType.upper() == "DOUBLE":
            if self._description[col][1] == FieldType.C_DOUBLE:
                return True
        if dataType.upper() == "BINARY" or dataType.upper() == "VARCHAR":
            if self._description[col][1] == FieldType.C_BINARY:
                return True
        if dataType.upper() == "TIMESTAMP":
            if self._description[col][1] == FieldType.C_TIMESTAMP:
                return True
        if dataType.upper() == "NCHAR":
            if self._description[col][1] == FieldType.C_NCHAR:
                return True
        if dataType.upper() == "JSON":
            if self._description[col][1] == FieldType.C_JSON:
                return True
        if dataType.upper() == "VARBINARY":
            if self._description[col][1] == FieldType.C_VARBINARY:
                return True

        return False

    def fetchall_row(self):
        """Fetch all (remaining) rows of a query result, returning them as a sequence of sequences (e.g. a list of tuples). Note that the cursor's arraysize attribute can affect the performance of this operation."""
        if self._result is None or self._fields is None:
            raise OperationalError("Invalid use of fetchall")

        buffer = [[] for i in range(len(self._fields))]
        self._rowcount = 0
        while True:
            block, num_of_rows = taos_fetch_row(self._result, self._fields, decode_binary=self.decode_binary)
            errno = taos_errno(self._result)
            if errno != 0:
                raise ProgrammingError(taos_errstr(self._result), errno)
            if num_of_rows == 0:
                break
            self._rowcount += num_of_rows
            for i in range(len(self._fields)):
                buffer[i].extend(block[i])
        return list(map(tuple, zip(*buffer)))

    def fetchall(self):
        if self._stmt_result is not None:
            return self._stmt_result.fetch_all()
        else:
            return self._fetchall_sql()

    def _fetchall_sql(self):
        if self._result is None:
            raise OperationalError("Invalid use of fetchall")
        fields = self._fields if self._fields is not None else taos_fetch_fields(self._result)
        buffer = [[] for i in range(len(fields))]
        self._rowcount = 0
        while True:
            block, num_of_rows = taos_fetch_block(self._result, self._fields, decode_binary=self.decode_binary)
            errno = taos_errno(self._result)
            if errno != 0:
                raise ProgrammingError(taos_errstr(self._result), errno)
            if num_of_rows == 0:
                break
            self._rowcount += num_of_rows
            for i in range(len(self._fields)):
                buffer[i].extend(block[i])
        return list(map(tuple, zip(*buffer)))

    def stop_query(self):
        if self._result is not None:
            taos_stop_query(self._result)

    def nextset(self):
        """ """
        pass

    def setinputsize(self, sizes):
        pass

    def setoutputsize(self, size, column=None):
        pass

    def _reset_result(self):
        """Reset the result to unused version."""
        self._description = []
        if self._result is not None:
            taos_free_result(self._result)
        self._result = None
        self._fields = None
        self._block = None
        self._block_rows = -1
        self._block_iter = 0
        self._affected_rows = 0

    def _handle_result(self):
        """Handle the return result from query."""
        self._description = []
        for ele in self._fields:
            self._description.append((ele["name"], ele["type"], None, None, None, None, False))

        return self._result

    def __del__(self):
        self.close()



from .connection import TaosRestConnection, Result
from .cursor import TaosRestCursor
from .restclient import RestClient
from .errors import *


def connect(**kwargs) -> TaosRestConnection:
    """
    Keyword Arguments
    ----------------------------
    - url: str, optional, default "http://localhost:6041"
         url to connect
    - token: str, optional, default None
         TDengine cloud Token, which is required only by TDengine cloud service
    - user : str, optional, default root
        username used to log in
    - password : str, optional, default taosdata
        password used to log in
    - database : str, optional, default None
         default database to use.
    - timeout : int, optional.
        the optional timeout parameter specifies a timeout in seconds for blocking operations
    - convert_timestamp: bool, optional, default true
         whether to convert timestamp in RFC3339 format to python datatime.
    - timezone: str | datetime.tzinfo, optional, default None.
         When convert_timestamp is true, which timezone to used.
         When the type of timezone is str, it should be recognized by [pytz package](https://pypi.org/project/pytz/).
         When the timezone is None, system timezone will be used and the returned datetime object will be offset-naive (no tzinfo), otherwise the returned datetime will be offset-aware(with tzinfo)
    Examples
    -----------------------------
     connect to cloud service
     ```python
     import taosrest, os
     url = os.environ("TDENGINE_CLOUD_URL")
     token = os.environ("TDENGINE_ClOUD_TOKEN")
     conn = taosrest.connect(url=url, token=token)
     ```
    connect to local taosAdapter
    ```python
    import taosrest
    conn = taosrest.connect(url="http://localhost:6041")
    ```
    """
    return TaosRestConnection(**kwargs)



# encoding:UTF-8
import sys
import ctypes
from ctypes import *
from datetime import datetime

from taos.cinterface import IS_V3
from taos.constants import FieldType
from taos.precision import PrecisionEnum, PrecisionError
from taos.bind_base import datetime_to_timestamp

_datetime_epoch = datetime.utcfromtimestamp(0)


class TaosBind(ctypes.Structure):
    _fields_ = [
        ("buffer_type", c_int),
        ("buffer", c_void_p),
        ("buffer_length", c_size_t),
        ("length", POINTER(c_size_t)),
        ("is_null", POINTER(c_int)),
        ("is_unsigned", c_int),
        ("error", POINTER(c_int)),
        ("u", c_int64),
        ("allocated", c_int),
    ]

    def bool(self, value):
        self.buffer_type = FieldType.C_BOOL
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_bool(value)), c_void_p)
            self.buffer_length = sizeof(c_bool)

    def tinyint(self, value):
        self.buffer_type = FieldType.C_TINYINT
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_int8(value)), c_void_p)
            self.buffer_length = sizeof(c_int8)

    def smallint(self, value):
        self.buffer_type = FieldType.C_SMALLINT
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_int16(value)), c_void_p)
            self.buffer_length = sizeof(c_int16)

    def int(self, value):
        self.buffer_type = FieldType.C_INT
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_int32(value)), c_void_p)
            self.buffer_length = sizeof(c_int32)

    def bigint(self, value):
        self.buffer_type = FieldType.C_BIGINT
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_int64(value)), c_void_p)
            self.buffer_length = sizeof(c_int64)

    def float(self, value):
        self.buffer_type = FieldType.C_FLOAT
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_float(value)), c_void_p)
            self.buffer_length = sizeof(c_float)

    def double(self, value):
        self.buffer_type = FieldType.C_DOUBLE
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_double(value)), c_void_p)
            self.buffer_length = sizeof(c_double)

    def binary(self, value):
        buffer = None
        length = 0
        self.buffer_type = FieldType.C_BINARY
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            if isinstance(value, str):
                bytes = value.encode("utf-8")
                buffer = create_string_buffer(bytes)
                length = len(bytes)
            else:
                buffer = value
                length = len(value)
            self.buffer = cast(buffer, c_void_p)
            self.buffer_length = length
            self.length = pointer(c_size_t(self.buffer_length))

    def timestamp(self, value, precision=PrecisionEnum.Milliseconds):
        self.buffer_type = FieldType.C_TIMESTAMP
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            if type(value) is datetime:
                if precision == PrecisionEnum.Milliseconds:
                    ts = int(round((value - _datetime_epoch).total_seconds() * 1000))
                elif precision == PrecisionEnum.Microseconds:
                    ts = int(round((value - _datetime_epoch).total_seconds() * 1000000))
                else:
                    raise PrecisionError("datetime do not support nanosecond precision")
            elif type(value) is float:
                if precision == PrecisionEnum.Milliseconds:
                    ts = int(round(value * 1000))
                elif precision == PrecisionEnum.Microseconds:
                    ts = int(round(value * 1000000))
                else:
                    raise PrecisionError("time float do not support nanosecond precision")
            elif isinstance(value, int) and not isinstance(value, bool):
                ts = value
            elif isinstance(value, str):
                value = datetime.fromisoformat(value)
                if precision == PrecisionEnum.Milliseconds:
                    ts = int(round((value - _datetime_epoch).total_seconds() * 1000))
                elif precision == PrecisionEnum.Microseconds:
                    ts = int(round((value - _datetime_epoch).total_seconds() * 1000000))
                else:
                    raise PrecisionError("datetime do not support nanosecond precision")
            self.buffer = cast(pointer(c_int64(ts)), c_void_p)
            self.buffer_length = sizeof(c_int64)

    def nchar(self, value):
        buffer = None
        length = 0
        self.buffer_type = FieldType.C_NCHAR
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            if isinstance(value, str):
                bytes = value.encode("utf-8")
                buffer = create_string_buffer(bytes)
                length = len(bytes)
            else:
                buffer = value
                length = len(value)
            self.buffer = cast(buffer, c_void_p)
            self.buffer_length = length
            self.length = pointer(c_size_t(self.buffer_length))

    def json(self, value):
        buffer = None
        length = 0
        self.buffer_type = FieldType.C_JSON
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            if isinstance(value, str):
                bytes = value.encode("utf-8")
                buffer = create_string_buffer(bytes)
                length = len(bytes)
            else:
                buffer = value
                length = len(value)
            self.buffer = cast(buffer, c_void_p)
            self.buffer_length = length
            self.length = pointer(c_size_t(self.buffer_length))

    def tinyint_unsigned(self, value):
        self.buffer_type = FieldType.C_TINYINT_UNSIGNED
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_uint8(value)), c_void_p)
            self.buffer_length = sizeof(c_uint8)

    def smallint_unsigned(self, value):
        self.buffer_type = FieldType.C_SMALLINT_UNSIGNED
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_uint16(value)), c_void_p)
            self.buffer_length = sizeof(c_uint16)

    def int_unsigned(self, value):
        self.buffer_type = FieldType.C_INT_UNSIGNED
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_uint32(value)), c_void_p)
            self.buffer_length = sizeof(c_uint32)

    def bigint_unsigned(self, value):
        self.buffer_type = FieldType.C_BIGINT_UNSIGNED
        if value is None:
            self.is_null = pointer(c_int(1))
        else:
            self.buffer = cast(pointer(c_uint64(value)), c_void_p)
            self.buffer_length = sizeof(c_uint64)

    def varchar(self, value):
        self.binary(value)


class TaosMultiBind(ctypes.Structure):
    _fields_ = [
        ("buffer_type", c_int),
        ("buffer", c_void_p),
        ("buffer_length", c_size_t),
        ("length", POINTER(c_int32)),
        ("is_null", c_char_p),
        ("num", c_int),
    ]

    def bool(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_int8 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_BOOL_NULL for v in values])

        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.buffer_type = FieldType.C_BOOL
        self.buffer_length = sizeof(c_bool)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def tinyint(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_TINYINT
        self.buffer_length = sizeof(c_int8)
        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_int8 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_TINYINT_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def smallint(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_SMALLINT
        self.buffer_length = sizeof(c_int16)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_int16 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_SMALLINT_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def int(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_INT
        self.buffer_length = sizeof(c_int32)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_int32 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_INT_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def bigint(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_BIGINT
        self.buffer_length = sizeof(c_int64)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_int64 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_BIGINT_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def float(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_FLOAT
        self.buffer_length = sizeof(c_float)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_float * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_FLOAT_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def double(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_DOUBLE
        self.buffer_length = sizeof(c_double)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_double * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_DOUBLE_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def _str_to_buffer(self, values, encode=True):
        self.num = len(values)
        is_null = [1 if v is None else 0 for v in values]
        self.is_null = cast((c_byte * self.num)(*is_null), c_char_p)

        if sum(is_null) == self.num:
            self.length = (c_int32 * len(values))(0 * self.num)
            return
        if sys.version_info < (3, 0):
            _bytes = [bytes(value) if value is not None else None for value in values]
            buffer_length = max(len(b) + 1 for b in _bytes if b is not None)
            buffers = [
                create_string_buffer(b, buffer_length) if b is not None else create_string_buffer(buffer_length)
                for b in _bytes
            ]
            buffer_all = b"".join(v[:] for v in buffers)
            self.buffer = cast(c_char_p(buffer_all), c_void_p)
        else:
            _bytes = []
            if encode:
                _bytes = [value.encode("utf-8") if value is not None else None for value in values]
            else:
                _bytes = [bytes(value) if value is not None else None for value in values]

            buffer_length = max(len(b) for b in _bytes if b is not None)
            self.buffer = cast(
                c_char_p(
                    b"".join(
                        [
                            (
                                create_string_buffer(b, buffer_length)
                                if b is not None
                                else create_string_buffer(buffer_length)
                            )
                            for b in _bytes
                        ]
                    )
                ),
                c_void_p,
            )
        self.length = (c_int32 * len(values))(*[len(b) if b is not None else 0 for b in _bytes])
        self.buffer_length = buffer_length

    def binary(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_BINARY
        self._str_to_buffer(values)

    def timestamp(self, values, precision=PrecisionEnum.Milliseconds):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_int64 * len(values)
            buffer = buffer_type(*[datetime_to_timestamp(value, precision) for value in values])

        self.buffer_type = FieldType.C_TIMESTAMP
        self.buffer = cast(buffer, c_void_p)
        self.buffer_length = sizeof(c_int64)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def nchar(self, values):
        # type: (list[str]) -> None
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_NCHAR
        self._str_to_buffer(values)

    def json(self, values):
        # type: (list[str]) -> None
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_JSON
        self._str_to_buffer(values)

    def tinyint_unsigned(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_TINYINT_UNSIGNED
        self.buffer_length = sizeof(c_uint8)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_uint8 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_TINYINT_UNSIGNED_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def smallint_unsigned(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_SMALLINT_UNSIGNED
        self.buffer_length = sizeof(c_uint16)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_uint16 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_SMALLINT_UNSIGNED_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def int_unsigned(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_INT_UNSIGNED
        self.buffer_length = sizeof(c_uint32)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_uint32 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_INT_UNSIGNED_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def bigint_unsigned(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_BIGINT_UNSIGNED
        self.buffer_length = sizeof(c_uint64)

        try:
            buffer = cast(values, c_void_p)
        except:
            buffer_type = c_uint64 * len(values)
            try:
                buffer = buffer_type(*values)
            except:
                buffer = buffer_type(*[v if v is not None else FieldType.C_BIGINT_UNSIGNED_NULL for v in values])
        self.buffer = cast(buffer, c_void_p)
        self.num = len(values)
        self.is_null = cast((c_char * len(values))(*[1 if value is None else 0 for value in values]), c_char_p)

    def varchar(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_VARCHAR
        self._str_to_buffer(values)

    def varbinary(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_VARBINARY
        self._str_to_buffer(values, False)

    def geometry(self, values):
        if type(values) is not tuple and type(values) is not list:
            values = tuple([values])
        self.buffer_type = FieldType.C_GEOMETRY
        self._str_to_buffer(values, False)


def new_bind_param():
    # type: () -> TaosBind
    if IS_V3:
        return TaosMultiBind()
    else:
        return TaosBind()


def new_bind_params(size):
    # type: (int) -> Array[TaosBind]
    if IS_V3:
        return (TaosMultiBind * size)()
    else:
        return (TaosBind * size)()


def new_multi_bind():
    # type: () -> TaosMultiBind
    return TaosMultiBind()


def new_multi_binds(size):
    # type: (int) -> Array[TaosMultiBind]
    return (TaosMultiBind * size)()



# encoding:UTF-8
from taos.cinterface import *
from taos.error import StatementError
from taos.result import TaosResult
from taos import bind2
from taos import log
from taos import utils
from taos.precision import PrecisionEnum

_taos_async_fn_t = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int)


#
# Bind data with table
#
class BindTable(object):
    def __init__(self, name, tags):
        self.name = name
        self.tags = tags
        self.datas = []

    def add_col_data(self, data):
        """Add column data to the table"""
        self.datas.append(data)


class TaosStmt2OptionImpl(ctypes.Structure):
    _fields_ = [
        ("reqid", ctypes.c_int64),
        ("singleStbInsert", ctypes.c_bool),
        ("singleTableBindOnce", ctypes.c_bool),
        ("asyncExecFn", _taos_async_fn_t),
        ("userdata", ctypes.c_void_p),
    ]


def taos_stmt2_async_exec(userdata, result_set, error_code):
    print(f"Executing asynchronously with userdata: {userdata}, result_set: {result_set}, error_code: {error_code}")


class TaosStmt2Option:
    def __init__(
        self, reqid: int = None, single_stb_insert: bool = False, single_table_bind_once: bool = False, **kwargs
    ):
        self._impl = TaosStmt2OptionImpl()
        if reqid is None:
            reqid = utils.gen_req_id()
        else:
            if type(reqid) is not int:
                raise StatementError(f"reqid type error, expected int type but got {type(reqid)}.")

        self.reqid = reqid
        self.single_stb_insert = single_stb_insert
        self.single_table_bind_once = single_table_bind_once
        self.is_async = kwargs.get("is_async", False)
        self.async_exec_fn = taos_stmt2_async_exec if self.is_async else _taos_async_fn_t()
        self.userdata = ctypes.c_void_p(None)

    @property
    def reqid(self) -> int:
        return self._impl.reqid

    @reqid.setter
    def reqid(self, value: int):
        self._impl.reqid = ctypes.c_int64(value)

    @property
    def single_stb_insert(self) -> bool:
        return self._impl.singleStbInsert

    @single_stb_insert.setter
    def single_stb_insert(self, value: bool):
        self._impl.singleStbInsert = ctypes.c_bool(value)

    @property
    def single_table_bind_once(self) -> bool:
        return self._impl.singleTableBindOnce

    @single_table_bind_once.setter
    def single_table_bind_once(self, value: bool):
        self._impl.singleTableBindOnce = ctypes.c_bool(value)

    @property
    def async_exec_fn(self) -> _taos_async_fn_t:
        return self._impl.asyncExecFn

    @async_exec_fn.setter
    def async_exec_fn(self, value: _taos_async_fn_t):
        if value is not None and not isinstance(value, _taos_async_fn_t):
            raise TypeError("async_exec_fn must be an instance of _taos_async_fn_t or None")
        self._impl.asyncExecFn = value

    @property
    def userdata(self) -> ctypes.c_void_p:
        return self._impl.userdata

    @userdata.setter
    def userdata(self, value: ctypes.c_void_p):
        self._impl.userdata = value

    def get_impl(self):
        return self._impl


#
#    ------------- global fun define -------------
#


def checkConsistent(tbnames, tags, datas):
    # todo
    return True


def obtainSchema(statement2):
    """Obtain schema information from statement2 object"""
    if statement2._stmt2 is None:
        raise StatementError("stmt2 object is null.")

    try:
        _, fields = statement2.get_fields()
        statement2.tag_fields = []
        statement2.fields = []
        statement2.all_fields = fields
        statement2.is_tbname = False
        for field in fields:
            if field.field_type == TAOS_FIELD_TAG:
                statement2.tag_fields.append(field)
            elif field.field_type == TAOS_FIELD_COL:
                statement2.fields.append(field)
            elif field.field_type == TAOS_FIELD_TBNAME:
                statement2.is_tbname = True
        log.debug(f"obtain schema tag fields = {statement2.tag_fields}")
        log.debug(f"obtain schema fields     = {statement2.fields}")
    except Exception as err:
        log.debug(f"obtain schema tag/col fields failed, reason: {repr(err)}")
        return False

    return len(statement2.fields) > 0


def getField(statement2, index, isTag):
    """Get field information by index and type"""
    if isTag:
        return statement2.tag_fields[index]
    else:
        return statement2.fields[index]


def createTagsBind(statement2, tags_tables):
    """Create stmt2Bind list from tags"""
    binds = []
    # Process tables tags
    for tags_table in tags_tables:
        # Process single table
        n = len(tags_table)
        binds_table = bind2.new_stmt2_binds(n)
        for i in range(n):
            field = getField(statement2, i, True)
            values = [tags_table[i]]
            log.debug(
                f"tag i = {i} type={field.type} precision={field.precision} length={field.bytes} values = {values} tags_table = {tags_table}\n"
            )
            binds_table[i].set_value(field.type, values, field.precision)
        binds.append(binds_table)

    return binds


def createColsBind(statement2, cols_tables):
    """Create stmt2Bind list from columns"""
    binds = []
    # Process tables columns data
    for cols_table in cols_tables:
        # Process single table
        n = len(cols_table)
        binds_table = bind2.new_stmt2_binds(n)
        for i in range(n):
            field = getField(statement2, i, isTag=False)
            binds_table[i].set_value(field.type, cols_table[i], field.precision)
        binds.append(binds_table)

    return binds


def createSuperBindV(statement2, cols_tables):
    """Create super table bind for automatic table name and tag binding"""
    if cols_tables is None:
        raise StatementError("data bind params is None.")
    bind_names = []
    bind_tags = []
    bind_cols = []
    tag_count = len(statement2.tag_fields)
    cols_count = len(statement2.all_fields)
    for cols_table in cols_tables:
        cols_bind = bind2.new_stmt2_binds(cols_count)
        tags_bind = bind2.new_stmt2_binds(tag_count)
        tag_index = 0
        cols_index = 0
        for i in range(cols_count):
            field = statement2.all_fields[i]

            log.debug(
                f"index i = {i} type={field.type} precision={field.precision} length={field.bytes} cols_table = {cols_table[i]}\n"
            )

            if field.field_type == TAOS_FIELD_TAG:
                values = [cols_table[i]]
                tags_bind[tag_index].set_value(field.type, values, field.precision)
                tag_index += 1
            elif field.field_type == TAOS_FIELD_COL:
                values = [cols_table[i]]
                cols_bind[cols_index].set_value(field.type, values, field.precision)
                cols_index += 1
            elif field.field_type == TAOS_FIELD_TBNAME:
                bind_names.append(cols_table[i])

        bind_tags.append(tags_bind)
        bind_cols.append(cols_bind)

    return bind2.new_bindv(len(bind_names), bind_names, bind_tags, bind_cols)


def createQueryBindV(statement2, datas):
    """Create bind for query operations"""
    if statement2._stmt2 is None:
        raise StatementError("stmt2 object is null.")

    # Create bindv
    ret = utils.detectListNone(datas)
    if ret == utils.ALL_NONE or ret == utils.HAVE_NONE:
        raise StatementError("params datas some is None, some is not None, this is error.")

    types = []
    query_array = []
    # Process tables columns data
    for cols_table in datas:
        n = len(cols_table)
        for i in range(n):
            value = cols_table[i]
            if isinstance(value, (list, tuple, set)):
                value = value[0]
            else:
                query_array.append([value])
            if isinstance(value, bool):
                types.append(FieldType.C_BOOL)
            elif isinstance(value, int):
                if value > 0:
                    types.append(FieldType.C_BIGINT_UNSIGNED)
                else:
                    types.append(FieldType.C_BIGINT)
            elif isinstance(value, float):
                types.append(FieldType.C_DOUBLE)
            elif isinstance(value, str):
                types.append(FieldType.C_BINARY)
            else:
                raise StatementError(
                    f"data type not support, only support int/float/str/bool type, but got {type(value)}"
                )

            log.debug(f"createQueryBindV cols_table={cols_table[i]} {type(cols_table[i])}")

    if len(query_array) > 0:
        datas = [query_array]

    statement2.set_columns_type(types)
    return createBindV(statement2, None, None, datas)


#
# Create bindv from list
#
def createBindV(statement2, tbnames, tags, datas):
    """Create bind vector from table names, tags, and data"""
    if tbnames is None and tags is None and datas is None:
        raise StatementError("all bind params is None.")

    log.debug(f"createBindV datas={datas}")

    # Count validation
    count = -1
    ret = utils.detectListNone(tbnames)
    if ret == utils.ALL_NONE:
        bind_names = None
    elif ret == utils.HAVE_NONE:
        raise StatementError("params tbnames some is None, some is not None, this is error.")
    else:
        # not found none
        if type(tbnames) not in [list, tuple]:
            raise StatementError(f"tbnames type error, expected list or tuple type but got {type(tbnames)}.")

        bind_names = tbnames
        count = len(tbnames)

    # Process tags
    ret = utils.detectListNone(tags)
    if ret == utils.ALL_NONE:
        bind_tags = None
    elif ret == utils.HAVE_NONE:
        raise StatementError("params tags some is None, some is not None, this is error.")
    else:
        # not found none
        bind_tags = createTagsBind(statement2, tags)
        if count == -1:
            count = len(tags)
        else:
            if count != len(tags):
                err = f"tags count is inconsistent. require count={count} real={len(tags)}"
                raise StatementError(err)

    # Process datas
    ret = utils.detectListNone(datas)
    if ret == utils.ALL_NONE:
        bind_datas = None
    elif ret == utils.HAVE_NONE:
        raise StatementError("params datas some is None, some is not None, this is error.")
    else:
        # not found none
        bind_datas = createColsBind(statement2, datas)
        if count == -1:
            count = len(datas)
        else:
            if count != len(datas):
                err = f"datas count is inconsistent. require count={count} real={len(datas)}"
                raise StatementError(err)

    # Create bind vector
    log.debug(f"createBindV count={count} bind_names={bind_names} bind_tags={bind_tags} bind_datas={bind_datas}")
    return bind2.new_bindv(count, bind_names, bind_tags, bind_datas)


#
# STMT2 object implementation
#


class TaosStmt2(object):
    """TDengine STMT2 interface for prepared statements"""

    def __init__(self, stmt2, decode_binary=True):
        self._stmt2 = stmt2
        self._decode_binary = decode_binary
        self.fields = None
        self.tag_fields = None
        self.all_fields = None
        self.is_tbname = False
        self.types = None
        self._is_insert = None
        self._affected_rows = 0

    def prepare(self, sql):
        if self._stmt2 is None:
            raise StatementError(ErrMsg.STMT2_NULL)

        if sql is None:
            raise StatementError("sql is null.")

        if not isinstance(sql, str):
            raise StatementError("sql is not str type.")

        if len(sql) == 0:
            raise StatementError("sql is empty.")

        self.fields = None
        self.tag_fields = None
        self._is_insert = None
        self.is_tbname = False

        taos_stmt2_prepare(self._stmt2, sql)

        # obtain schema if insert
        if self.is_insert():
            if obtainSchema(self) is False:
                raise StatementError(f"obtain schema failed, maybe this sql is not supported, sql: {sql}")

    def bind_param(self, tbnames, tags, datas):
        if self._stmt2 is None:
            raise StatementError(ErrMsg.STMT2_NULL)

        log.debug(f"bind_param tbnames  = {tbnames} \n")
        log.debug(f"bind_param tags     = {tags}    \n")
        log.debug(f"bind_param datasTbs = {datas}   \n")

        # check consistent
        if checkConsistent(tbnames, tags, datas) == False:
            raise StatementError("check consistent failed.")

        # bindV
        bindv = None
        if self._is_insert and self.is_tbname and tbnames is None and tags is None:
            log.debug("insert with super bind mode.\n")
            bindv = createSuperBindV(self, datas)
        elif self._is_insert == False:
            bindv = createQueryBindV(self, datas)
        else:
            bindv = createBindV(self, tbnames, tags, datas)

        if bindv is None:
            raise StatementError("create stmt2 bindV failed.")

        # Call engine
        taos_stmt2_bind_param(self._stmt2, bindv.get_address(), -1)

    def bind_param_with_tables(self, tables):
        """Bind parameters using table objects"""
        tbnames = []
        tags_tables = []
        datas_tables = []

        for table in tables:
            tbnames.append(table.name)
            tags_tables.append(table.tags)
            datas_tables.append(table.datas)

        # Call bind
        self.bind_param(tbnames, tags_tables, datas_tables)

    def execute(self) -> int:
        """Execute the prepared statement"""
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")

        self._affected_rows = taos_stmt2_exec(self._stmt2)
        return self._affected_rows

    def result(self):
        """Get result from the executed statement"""
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")

        result = taos_stmt2_result(self._stmt2)
        return TaosResult(result, close_after=False, decode_binary=self._decode_binary)

    def close(self):
        """Close the statement and release resources"""
        if self._stmt2 is None:
            return

        taos_stmt2_close(self._stmt2)
        self._stmt2 = None

    def is_insert(self):
        """Check if the statement is an INSERT statement"""
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")

        if self._is_insert is None:
            self._is_insert = taos_stmt2_is_insert(self._stmt2)

        return self._is_insert

    def get_fields(self):
        """Get field information from the statement"""
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")

        return taos_stmt2_get_fields(self._stmt2)

    def error(self):
        """Get error information from the statement"""
        if self._stmt2 is None:
            raise StatementError("stmt2 object is null.")

        return taos_stmt2_error(self._stmt2)

    def set_columns_type(self, types):
        """Set columns type for query operations

        If query must set columns type (not engine interface),
        type is defined in class FieldType in constants.py
        """
        self.fields = []
        for field_type in types:
            item = TaosFieldAllCls(None, field_type, PrecisionEnum.Milliseconds, None, None, TAOS_FIELD_COL)
            self.fields.append(item)

    def __del__(self):
        self.close()

    @property
    def affected_rows(self):
        # type: () -> int
        return self._affected_rows



import sys
from sqlalchemy import types as sqltypes
from sqlalchemy.engine import default, reflection
from sqlalchemy import text

# taos.sqlalchemy.BaseDialect
import taos

#
# ---------------- taos rest impl -------------
#
import taos.sqlalchemy
import taosrest

# rest connect
class AlchemyRestConnection:
    threadsafety = 0
    paramstyle = "pyformat"
    Error = taosrest.Error
    # connect
    def connect(self, **kwargs):
        host = kwargs["host"] if "host" in kwargs else None
        port = kwargs["port"] if "port" in kwargs else None
        user = kwargs["username"] if "username" in kwargs else "root"
        password = kwargs["password"] if "password" in kwargs else "taosdata"
        database = kwargs["database"] if "database" in kwargs else None
        token = kwargs["token"] if "token" in kwargs else None

        if not host:
            host = "localhost"
        if host == "localhost" and not port:
            port = 6041

        url = f"http://{host}"
        if port:
            url += f":{port}"
        return taosrest.connect(url=url, user=user, password=password, database=database, token=token)


# rest dialect
class TaosRestDialect(taos.sqlalchemy.BaseDialect):
    name = "taosrest"
    driver = "taosrest"

    @classmethod
    def dbapi(cls):
        return AlchemyRestConnection()

    @classmethod
    def import_dbapi(cls):
        return AlchemyRestConnection()



class PrecisionEnum(object):
    """Precision enums"""

    Milliseconds = 0
    Microseconds = 1
    Nanoseconds = 2


class PrecisionError(Exception):
    """Python datetime does not support nanoseconds error"""

    pass



class SmlPrecision:
    """Schemaless timestamp precision constants"""

    NOT_CONFIGURED = 0  # C.TSDB_SML_TIMESTAMP_NOT_CONFIGURED
    HOURS = 1
    MINUTES = 2
    SECONDS = 3
    MILLI_SECONDS = 4
    MICRO_SECONDS = 5
    NANO_SECONDS = 6


class SmlProtocol:
    """Schemaless protocol constants"""

    UNKNOWN_PROTOCOL = 0
    LINE_PROTOCOL = 1
    TELNET_PROTOCOL = 2
    JSON_PROTOCOL = 3



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['taos', 'taosrest']

package_data = \
{'': ['*']}

install_requires = \
['iso8601==1.0.2',
 'pytz',
 'requests>=2.27.1',
 'typing-extensions>=4.2.0,<4.15.0']

extras_require = \
{'ws:python_version >= "3.7" and python_version < "4.0"': ['taos-ws-py>=0.3.0']}

entry_points = \
{'sqlalchemy.dialects': ['taos = taos.sqlalchemy:TaosDialect',
                         'taosrest = taosrest.sqlalchemy:TaosRestDialect',
                         'taosws = taos.sqlalchemy:TaosWsDialect']}

setup_kwargs = {
    'name': 'taospy',
    'version': '2.8.8',
    'description': 'The official TDengine Python connector',
    'long_description': '<!-- omit in toc -->\n# TDengine Python Connector\n\n\n[![GitHub Actions Workflow Status](https://img.shields.io/github/actions/workflow/status/taosdata/taos-connector-python/build.yml)](https://github.com/taosdata/taos-connector-python/actions/workflows/build.yml)\n[![codecov](https://codecov.io/gh/taosdata/taos-connector-python/branch/main/graph/badge.svg?token=BDANN3DBXS)](https://codecov.io/gh/taosdata/taos-connector-python)\n![GitHub commit activity](https://img.shields.io/github/commit-activity/m/taosdata/taos-connector-python)\n![PyPI](https://img.shields.io/pypi/dm/taospy)\n![GitHub License](https://img.shields.io/github/license/taosdata/taos-connector-python)\n[![PyPI](https://img.shields.io/pypi/v/taospy)](https://pypi.org/project/taospy/)\n<br />\n[![Twitter Follow](https://img.shields.io/twitter/follow/tdenginedb?label=TDengine&style=social)](https://twitter.com/tdenginedb)\n[![YouTube Channel](https://img.shields.io/badge/Subscribe_@tdengine--white?logo=youtube&style=social)](https://www.youtube.com/@tdengine)\n[![Discord Community](https://img.shields.io/badge/Join_Discord--white?logo=discord&style=social)](https://discord.com/invite/VZdSuUg4pS)\n[![LinkedIn](https://img.shields.io/badge/Follow_LinkedIn--white?logo=linkedin&style=social)](https://www.linkedin.com/company/tdengine)\n[![StackOverflow](https://img.shields.io/badge/Ask_StackOverflow--white?logo=stackoverflow&style=social&logoColor=orange)](https://stackoverflow.com/questions/tagged/tdengine)\n\nEnglish | [简体中文](./README-CN.md)\n\n<!-- omit in toc -->\n## Table of Contents\n\n- [1. Introduction](#1-introduction)\n- [2. Documentation](#2-documentation)\n- [3. Prerequisites](#3-prerequisites)\n- [4. Build](#4-build)\n- [5. Testing](#5-testing)\n  - [5.1 Test Execution](#51-test-execution)\n  - [5.2 Test Case Addition](#52-test-case-addition)\n  - [5.3 Performance Testing](#53-performance-testing)\n- [6. CI/CD](#6-cicd)\n- [7. Submitting Issues](#7-submitting-issues)\n- [8. Submitting PRs](#8-submitting-prs)\n- [9. References](#9-references)\n- [10. License](#10-license)\n\n## 1. Introduction\n\n`taospy` is the official Python Connector for TDengine, allowing Python developers to develop applications that access the TDengine database. It supports functions such as data writing, querying, subscription, schemaless writing, and parameter binding.\n\nThe API for `taospy` is compliant with the Python DB API 2.0 (PEP-249). It contains two modules:\n\n1. The `taos` module. It uses TDengine C client library for client server communications.\n2. The `taosrest` module. It wraps TDengine RESTful API to Python DB API 2.0 (PEP-249). With this module, you do not need to install the TDengine C client library.\n\n## 2. Documentation\n\n- To use Python Connector, please check [Developer Guide](https://docs.tdengine.com/developer-guide/), which includes how an application can introduce the Python Connector , as well as examples of data writing, querying, schemaless writing, parameter binding, and data subscription.\n- For other reference information, please check [Reference Manual](https://docs.taosdata.com/reference/connector/python/), which includes version history, data types, example programs, API descriptions, and FAQs.\n- This quick guide is mainly for developers who like to contribute/build/test the Python Connector by themselves. To learn about TDengine, you can visit the [official documentation](https://docs.tdengine.com).\n\n## 3. Prerequisites\n\n- Python runtime environment (taospy: Python >= 3.6.2, taos-ws-py: Python >= 3.7)\n- TDengine has been deployed locally. For specific steps, please refer to [Deploy TDengine](https://docs.tdengine.com/get-started/deploy-from-package/), and taosd and taosAdapter have been started.\n\n## 4. Build\n\nDownload the repository code and execute the following in root directory to build develop environment:\n``` bash\npip3 install -e ./ \n```\n\n## 5. Testing\n### 5.1 Test Execution\nThe Python Connector testing framework is `pytest`  \nThe testing directory for `taospy` is located in the root directory: tests/  \nThe testing directory for `taos-ws-py` is located in the root directory: taos-ws-py/tests/  \nThe test code has been written into one bash file. You can open and view the detailed testing process   \nThe following command runs all test cases on Linux platform:\n``` bash\n# for taospy\nbash ./test_taospy.sh\n```\n\n``` bash\n# for taos-ws-py\nbash ./test_taos-ws-py.sh\n```\n\n### 5.2 Test Case Addition\nYou can add new test files or add test cases in existing test files that comply with `pytest` standards\n\n### 5.3 Performance Testing\nPerformance testing is in progress.\n\n## 6. CI/CD\n- [Build Workflow](https://github.com/taosdata/taos-connector-python/actions/workflows/build.yml)\n- [Code Coverage](https://app.codecov.io/gh/taosdata/taos-connector-python)\n\n## 7. Submitting Issues\nWe welcome the submission of [GitHub Issue](https://github.com/taosdata/taos-connector-python/issues/new?template=Blank+issue). When submitting, please provide the following information:\n\n- Problem description, whether it always occurs, and it\'s best to include a detailed call stack.\n- Python Connector version.\n- Python Connection parameters (username and password not required).\n- TDengine server version.\n\n## 8. Submitting PRs\nWe welcome developers to contribute to this project. When submitting PRs, please follow these steps:\n\n1. Fork this project, refer to ([how to fork a repo](https://docs.github.com/en/get-started/quickstart/fork-a-repo)).\n2. Create a new branch from the main branch with a meaningful branch name (`git checkout -b my_branch`). Do not modify the main branch directly.\n3. Modify the code, ensure all unit tests pass, and add new unit tests to verify the changes.\n4. Push the changes to the remote branch (`git push origin my_branch`).\n5. Create a Pull Request on GitHub ([how to create a pull request](https://docs.github.com/en/pull-requests/collaborating-with-pull-requests/proposing-changes-to-your-work-with-pull-requests/creating-a-pull-request)).\n6. After submitting the PR, you can find your PR through the [Pull Request](https://github.com/taosdata/taos-connector-python/pulls). Click on the corresponding link to see if the CI for your PR has passed. If it has passed, it will display "All checks have passed". Regardless of whether the CI passes or not, you can click "Show all checks" -> "Details" to view the detailed test case logs.\n7. After submitting the PR, if CI passes, you can find your PR on the [codecov](https://app.codecov.io/gh/taosdata/taos-connector-python/pulls) page to check the test coverage.\n\n## 9. References\n- [TDengine Official Website](https://www.tdengine.com/) \n- [TDengine GitHub](https://github.com/taosdata/TDengine) \n\n## 10. License\n[MIT License](./LICENSE)',
    'author': 'Taosdata Inc.',
    'author_email': 'support@taosdata.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# encoding:UTF-8
from taos.bind import *
from taos.bind2 import *
from taos.connection import TaosConnection
from taos.cursor import *

# For some reason, the following is needed for VS Code (through PyLance) to
# recognize that "error" is a valid module of the "taos" package.
from taos.error import *
from taos.field import *
from taos.result import *
from taos.schemaless import *
from taos.statement import *
from taos.statement2 import *
from taos.subscription import *

try:
    from taos.sqlalchemy import *
except:
    pass

from taos._version import __version__

# Globals
threadsafety = 2
"""sqlalchemy will read this attribute"""
paramstyle = "qmark"
"""sqlalchemy will read this attribute"""

__all__ = [
    "__version__",
    "IS_V3",
    "IGNORE",
    "TAOS_FIELD_COL",
    "TAOS_FIELD_TAG",
    "TAOS_FIELD_QUERY",
    "TAOS_FIELD_TBNAME",
    # functions
    "connect",
    "connect_test",
    "new_bind_param",
    "new_bind_params",
    "new_multi_binds",
    "new_multi_bind",
    # objects
    "TaosBind",
    "TaosConnection",
    "TaosCursor",
    "TaosResult",
    "TaosRows",
    "TaosRow",
    "TaosStmt",
    "TaosStmt2",
    "TaosStmt2Option",
    "BindTable",
    "PrecisionEnum",
    "SmlPrecision",
    "SmlProtocol",
    "utils",
]


def connect(*args, **kwargs):
    # type: (..., ...) -> TaosConnection
    """Function to return a TDengine connector object

    Current supporting keyword parameters:
    @dsn: Data source name as string
    @host: Hostname(optional)
    @port: Port number(optional)
    @user: Username as string(optional)
    @password: Password as string(optional)
    @totp_code: TOTP code for TOTP authentication(optional)
    @bearer_token: Token for token authentication(optional)
    @database: Database name(optional)

    @rtype: TDengineConnector
    """
    return TaosConnection(*args, **kwargs)


def connect_test(**kwargs):
    # type: (...) -> None
    """Function to test connection to TDengine with TOTP authentication

    Current supporting keyword parameters:
    @host: Hostname(optional)
    @port: Port number(optional)
    @user: Username as string(optional)
    @password: Password as string(optional)
    @totp_code: TOTP code for TOTP authentication(required)
    @database: Database name(optional)
    """
    from taos.cinterface import taos_connect_test

    host = kwargs.get("host")
    port = kwargs.get("port", 0)
    user = kwargs.get("user", "root")
    password = kwargs.get("password", "taosdata")
    totp = kwargs.get("totp_code")
    db = kwargs.get("database")

    if totp is None:
        raise TypeError("connect_test() missing 1 required keyword-only argument: 'totp_code'")

    taos_connect_test(host, user, password, totp, db, port)



# encoding:UTF-8
from typing import Optional, List

from taos.cinterface import *
from taos.constants import TSDB_CONNECTIONS_MODE, TSDB_OPTION_CONNECTION
from taos.result import TaosResult
from taos.cursor import TaosCursor
from taos.subscription import TaosSubscription
from taos.statement import TaosStmt
from taos.statement2 import TaosStmt2, TaosStmt2Option


class TaosConnection(object):
    """TDengine connection object"""

    def __init__(self, *args, **kwargs):
        self._conn = None
        self._host = None
        self._user = "root"
        self._password = "taosdata"
        self._database = None
        self._port = 0
        self._config = None
        self._tz = None
        self.decode_binary = True
        self._charset = None
        self._user_app = None
        self._user_ip = None
        self._bi_mode = None
        self._totp_code = None
        self._bearer_token = None

        self._init_config(**kwargs)
        self._chandle = CTaosInterface(self._config, self._tz)

        if self._bearer_token:
            self._conn = self._chandle.connect_token(self._host, self._bearer_token, self._database, self._port)
        elif self._totp_code:
            self._conn = self._chandle.connect_totp(
                self._host, self._user, self._password, self._totp_code, self._database, self._port
            )
        else:
            self._conn = self._chandle.connect(self._host, self._user, self._password, self._database, self._port)

        if self._charset is not None:
            self.set_option(TSDB_OPTION_CONNECTION.TSDB_OPTION_CONNECTION_CHARSET.value, self._charset)

        if self._user_app is not None:
            self.set_option(TSDB_OPTION_CONNECTION.TSDB_OPTION_CONNECTION_USER_APP.value, self._user_app)

        if self._user_ip is not None:
            self.set_option(TSDB_OPTION_CONNECTION.TSDB_OPTION_CONNECTION_USER_IP.value, self._user_ip)

        if self._bi_mode is not None:
            if self._bi_mode:
                self.set_mode(TSDB_CONNECTIONS_MODE.TSDB_CONNECTIONS_MODE_BI.value, 1)
            else:
                self.set_mode(TSDB_CONNECTIONS_MODE.TSDB_CONNECTIONS_MODE_BI.value, 0)

    def _init_config(self, **kwargs):
        # host
        if "host" in kwargs:
            self._host = kwargs["host"]

        # user
        if "user" in kwargs:
            self._user = kwargs["user"]

        # password
        if "password" in kwargs:
            self._password = kwargs["password"]

        # database
        if "database" in kwargs:
            self._database = kwargs["database"]

        # port
        if "port" in kwargs:
            self._port = kwargs["port"]

        # config
        if "config" in kwargs:
            self._config = kwargs["config"]

        # timezone
        if "timezone" in kwargs:
            self._tz = kwargs["timezone"]

        if "decode_binary" in kwargs:
            self.decode_binary = kwargs["decode_binary"]

        if "charset" in kwargs:
            self._charset = kwargs["charset"]

        if "user_app" in kwargs:
            self._user_app = kwargs["user_app"]

        if "user_ip" in kwargs:
            self._user_ip = kwargs["user_ip"]

        if "bi_mode" in kwargs:
            self._bi_mode = kwargs["bi_mode"]

        if "totp_code" in kwargs:
            self._totp_code = kwargs["totp_code"]

        if "bearer_token" in kwargs:
            self._bearer_token = kwargs["bearer_token"]

    def close(self):
        """Close current connection."""
        if self._conn:
            taos_close(self._conn)
            self._conn = None

    @property
    def client_info(self):
        # type: () -> str
        return taos_get_client_info()

    @property
    def server_info(self):
        # type: () -> str
        return taos_get_server_info(self._conn)

    def set_option(self, option, value):
        taos_options_connection(self._conn, option, value)

    def set_mode(self, mode, value):
        taos_set_conn_mode(self._conn, mode, value)

    def select_db(self, database):
        # type: (str) -> None
        taos_select_db(self._conn, database)

    def execute(self, sql, req_id: Optional[int] = None):
        """Simplely execute sql ignoring the results"""
        return self.query(sql, req_id).affected_rows

    def query(self, sql: str, req_id: Optional[int] = None) -> TaosResult:
        if req_id is None:
            res = taos_query(self._conn, sql)
        else:
            res = taos_query_with_reqid(self._conn, sql, req_id)
        return TaosResult(res, close_after=True, decode_binary=self.decode_binary)

    def query_a(self, sql: str, callback: async_query_callback_type, param: c_void_p, req_id: Optional[int] = None):
        """Asynchronously query a sql with callback function"""
        if req_id is None:
            taos_query_a(self._conn, sql, callback, param)
        else:
            taos_query_a_with_reqid(self._conn, sql, callback, param, req_id)

    def subscribe(
        self,
        restart: bool,
        topic: str,
        sql: str,
        interval: int,
        callback: Optional[subscribe_callback_type] = None,
        param: Optional[c_void_p] = None,
    ) -> Optional[TaosSubscription]:
        """Create a subscription."""
        if self._conn is None:
            return None
        sub = taos_subscribe(self._conn, restart, topic, sql, c_int(interval), callback, param)
        if not sub:
            errno = taos_errno(c_void_p(None))
            msg = taos_errstr(c_void_p(None))
            raise Error(msg, errno)
        return TaosSubscription(sub, callback is not None)

    def statement(self, sql=None):
        # type: (str | None) -> TaosStmt|None
        if self._conn is None:
            return None
        stmt = taos_stmt_init(self._conn)
        if sql is not None:
            taos_stmt_prepare(stmt, sql)

        return TaosStmt(stmt, decode_binary=self.decode_binary)

    def statement2(self, sql=None, option=None):
        # type: (str | None, TaosStmt2Option | None) -> TaosStmt2|None
        if self._conn is None:
            return None
        if option is not None:
            option = option.get_impl()
        _stmt2 = taos_stmt2_init(self._conn, option)
        stmt2 = TaosStmt2(_stmt2, decode_binary=self.decode_binary)
        if sql is not None:
            stmt2.prepare(sql)

        return stmt2

    def load_table_info(self, tables):
        # type: (str) -> None
        taos_load_table_info(self._conn, tables)

    def schemaless_insert(
        self,
        lines: List[str],
        protocol: SmlProtocol,
        precision: SmlPrecision,
        req_id: Optional[int] = None,
        ttl: Optional[int] = None,
    ) -> int:
        """
        1.Line protocol and schemaless support

        ## Example

        ```python
        import taos
        conn = taos.connect()
        conn.exec("drop database if exists test")
        conn.select_db("test")
        lines = [
            'ste,t2=5,t3=L"ste" c1=true,c2=4,c3="string" 1626056811855516532',
        ]
        conn.schemaless_insert(lines, 0, "ns")
        ```

        2.OpenTSDB telnet style API format support

        ## Example

        ```python
        import taos
        conn = taos.connect()
        conn.exec("drop database if exists test")
        conn.select_db("test")
        lines = [
            'cpu_load 1626056811855516532ns 2.0f32 id="tb1",host="host0",interface="eth0"',
        ]
        conn.schemaless_insert(lines, 1, None)
        ```

        3.OpenTSDB HTTP JSON format support

        ## Example

        ```python
        import taos
        conn = taos.connect()
        conn.exec("drop database if exists test")
        conn.select_db("test")
        payload = ['''
        {
            "metric": "cpu_load_0",
            "timestamp": 1626006833610123,
            "value": 55.5,
            "tags":
                {
                    "host": "ubuntu",
                    "interface": "eth0",
                    "Id": "tb0"
                }
        }
        ''']
        conn.schemaless_insert(lines, 2, None)
        ```
        """
        if ttl is None:
            if req_id is None:
                return taos_schemaless_insert(
                    self._conn,
                    lines,
                    protocol,
                    precision,
                )
            else:
                return taos_schemaless_insert_with_reqid(
                    self._conn,
                    lines,
                    protocol,
                    precision,
                    req_id,
                )
        else:
            if req_id is None:
                return taos_schemaless_insert_ttl(
                    self._conn,
                    lines,
                    protocol,
                    precision,
                    ttl,
                )
            else:
                return taos_schemaless_insert_ttl_with_reqid(
                    self._conn,
                    lines,
                    protocol,
                    precision,
                    ttl,
                    req_id,
                )
                pass

    def schemaless_insert_raw(
        self,
        lines: str,
        protocol: SmlProtocol,
        precision: SmlPrecision,
        req_id: Optional[int] = None,
        ttl: Optional[int] = None,
    ) -> int:
        """
        1.Line protocol and schemaless support

        ## Example

        ```python
        import taos
        conn = taos.connect()
        conn.exec("drop database if exists test")
        conn.select_db("test")
        lines = 'ste,t2=5,t3=L"ste" c1=true,c2=4,c3="string" 1626056811855516532'
        conn.schemaless_insert_raw(lines, 0, "ns")
        ```

        2.OpenTSDB telnet style API format support

        ## Example

        ```python
        import taos
        conn = taos.connect()
        conn.exec("drop database if exists test")
        conn.select_db("test")
        lines = 'cpu_load 1626056811855516532ns 2.0f32 id="tb1",host="host0",interface="eth0"'
        conn.schemaless_insert_raw(lines, 1, None)
        ```

        3.OpenTSDB HTTP JSON format support

        ## Example

        ```python
        import taos
        conn = taos.connect()
        conn.exec("drop database if exists test")
        conn.select_db("test")
        payload = '''
        {
            "metric": "cpu_load_0",
            "timestamp": 1626006833610123,
            "value": 55.5,
            "tags":
                {
                    "host": "ubuntu",
                    "interface": "eth0",
                    "Id": "tb0"
                }
        }
        '''
        conn.schemaless_insert_raw(lines, 2, None)
        ```
        """
        if ttl is None:
            if req_id is None:
                return taos_schemaless_insert_raw(
                    self._conn,
                    lines,
                    protocol,
                    precision,
                )
            else:
                return taos_schemaless_insert_raw_with_reqid(
                    self._conn,
                    lines,
                    protocol,
                    precision,
                    req_id,
                )
        else:
            if req_id is None:
                return taos_schemaless_insert_raw_ttl(
                    self._conn,
                    lines,
                    protocol,
                    precision,
                    ttl,
                )
            else:
                return taos_schemaless_insert_raw_ttl_with_reqid(
                    self._conn,
                    lines,
                    protocol,
                    precision,
                    ttl,
                    req_id,
                )

    def cursor(self):
        # type: () -> TaosCursor
        """Return a new Cursor object using the connection."""
        return TaosCursor(self, decode_binary=self.decode_binary)

    def commit(self):
        """Commit any pending transaction to the database.

        Since TDengine do not support transactions, the implement is void functionality.
        """
        pass

    def rollback(self):
        """Void functionality"""
        pass

    def clear_result_set(self):
        """Clear unused result set on this connection."""
        pass

    def __del__(self):
        self.close()

    def get_table_vgroup_id(self, db, table):
        # type: (str, str) -> int
        """
        get table's vgroup id. It's require db name and table name, and return an int type vgroup id.
        """
        return taos_get_table_vgId(self._conn, db, table)


if __name__ == "__main__":
    conn = TaosConnection()
    conn.close()
